from .user_action_types import UserActionTypes


class UserAction:
    """This class will be updated in next versions."""

    def __init__(self, prp_id: int, action: UserActionTypes, prp_value_id: int, value_id: int, value: any, datatype: str,
                 multi: bool, part: int):
        self.prp_id = prp_id
        self.action = action
        self.prp_value_id = prp_value_id
        self.value_id = value_id
        self.value = value
        self.part = part
        self.datatype = datatype
        self.multi = multi

    def as_tuple(self) -> tuple:
        return (self.prp_id, self.action.value, self.prp_value_id, self.value_id,
                self.value, self.part, self.datatype, self.multi)

    def as_dict(self) -> dict:
        return {
            "prp_id": self.prp_id,
            "action": self.action.value,
            "prp_value_id": self.prp_value_id,
            "value_id": self.value_id,
            "value": self.value,
            "part": self.part,
            "datatype": self.datatype,
            "multi": self.multi
        }

    def __str__(self) -> str:
        return str(self.as_tuple())

    def __repr__(self) -> str:
        return str(self.as_dict())




from enum import Enum


class HttpStatusCodes(Enum):
    CONTINUE = "100 Continue"
    SWITCHING_PROTOCOLS = "101 Switching protocols"
    PROCESSING = "102 Processing"
    EARLY_HINTS = "103 Early Hints"
    OK = "200 OK"
    CREATED = "201 Created"
    ACCEPTED = "202 Accepted"
    NON_AUTHORITATIVE_INFORMATION = "203 Non-Authoritative Information"
    NO_CONTENT = "204 No Content"
    RESET_CONTENT = "205 Reset Content"
    PARTIAL_CONTENT = "206 Partial Content"
    MULTI_STATUS = "207 Multi-Status"
    ALREADY_REPORTED = "208 Already Reported"
    IM_USED = "226 IM Used"
    MULTIPLE_CHOICES = "300 Multiple Choices"
    MOVED_PERMANENTLY = "301 Moved Permanently"
    FOUND = "302 Found (Previously \"Moved temporarily\")"
    SEE_OTHER = "303 See Other"
    NOT_MODIFIED = "304 Not Modified"
    USE_PROXY = "305 Use Proxy"
    SWITCH_PROXY = "306 Switch Proxy"
    TEMPORARY_REDIRECT = "307 Temporary Redirect"
    PERMANENT_REDIRECT = "308 Permanent Redirect"
    BAD_REQUEST = "400 Bad Request"
    UNAUTHORIZED = "401 Unauthorized"
    PAYMENT_REQUIRED = "402 Payment Required"
    FORBIDDEN = "403 Forbidden"
    NOT_FOUND = "404 Not Found"
    METHOD_NOT_ALLOWED = "405 Method Not Allowed"
    NOT_ACCEPTABLE = "406 Not Acceptable"
    PROXY_AUTHENTICATION_REQUIRED = "407 Proxy Authentication Required"
    REQUEST_TIMEOUT = "408 Request Timeout"
    CONFLICT = "409 Conflict"
    GONE = "410 Gone"
    LENGTH_REQUIRED = "411 Length Required"
    PRECONDITION_FAILED = "412 Precondition Failed"
    PAYLOAD_TOO_LARGE = "413 Payload Too Large"
    URI_TOO_LONG = "414 URI Too Long"
    UNSUPPORTED_MEDIA_TYPE = "415 Unsupported Media Type"
    RANGE_NOT_SATISFIABLE = "416 Range Not Satisfiable"
    EXPECTATION_FAILED = "417 Expectation Failed"
    I_AM_A_TEAPOT = "418 I'm a Teapot"
    MISDIRECTED_REQUEST = "421 Misdirected Request"
    UNPROCESSABLE_ENTITY = "422 Unprocessable Entity"
    LOCKED = "423 Locked"
    FAILED_DEPENDENCY = "424 Failed Dependency"
    TOO_EARLY = "425 Too Early"
    UPGRADE_REQUIRED = "426 Upgrade Required"
    PRECONDITION_REQUIRED = "428 Precondition Required"
    TOO_MANY_REQUESTS = "429 Too Many Requests"
    REQUEST_HEADER_FIELDS_TOO_LARGE = "431 Request Header Fields Too Large"
    UNAVAILABLE_FOR_LEGAL_REASONS = "451 Unavailable For Legal Reasons"
    INTERNAL_SERVER_ERROR = "500 Internal Server Error"
    NOT_IMPLEMENTED = "501 Not Implemented"
    BAD_GATEWAY = "502 Bad Gateway"
    SERVICE_UNAVAILABLE = "503 Service Unavailable"
    GATEWAY_TIMEOUT = "504 Gateway Timeout"
    HTTP_VERSION_NOT_SUPPORTED = "505 HTTP Version Not Supported"
    VARIANT_ALSO_NEGOTIATES = "506 Variant Also Negotiates"
    INSUFFICIENT_STORAGE = "507 Insufficient Storage"
    LOOP_DETECTED = "508 Loop Detected"
    NOT_EXTENDED = "510 Not Extended"
    NETWORK_AUTHENTICATION_REQUIRED = "511 Network Authentication Required"



from abc import ABC
import traceback
from typing import TYPE_CHECKING, Tuple

from bclib.exception import ShortCircuitErr

if TYPE_CHECKING:
    from .. import dispatcher

from bclib.db_manager import SqlDb, SQLiteDb, MongoDb, RabbitConnection, RESTfulConnection
from bclib.utility import DictEx, HttpStatusCodes


class Context(ABC):
    """Base class for dispatching"""

    def __init__(self, dispatcher: 'dispatcher.IDispatcher') -> None:
        super().__init__()
        self.dispatcher = dispatcher
        self.url_segments: DictEx = None
        self.url: str = None
        self.is_adhoc = True

    def open_sql_connection(self, key: str) -> SqlDb:
        return self.dispatcher.db_manager.open_sql_connection(key)

    def open_sqllite_connection(self, key: str) -> SQLiteDb:
        return self.dispatcher.db_manager.open_sqllite_connection(key)

    def open_mongo_connection(self, key: str) -> MongoDb:
        return self.dispatcher.db_manager.open_mongo_connection(key)

    def open_restful_connection(self, key: str) -> RESTfulConnection:
        return self.dispatcher.db_manager.open_restful_connection(key)

    def open_rabbit_connection(self, key: str) -> RabbitConnection:
        return self.dispatcher.db_manager.open_rabbit_connection(key)

    def generate_error_response(self, exception: Exception) -> dict:
        """Generate error response from process result"""

        error_object, _ = self._generate_error_object(exception)
        return error_object

    def _generate_error_object(self, exception: Exception) -> 'Tuple[dict, HttpStatusCodes]':
        """Generate error object from exception object"""
        error_code = None
        status_code = HttpStatusCodes.INTERNAL_SERVER_ERROR
        if isinstance(exception, ShortCircuitErr):
            status_code = exception.status_code
            error_code = exception.error_code
        error_object = {
            "errorCode": error_code,
            "errorMessage": str(exception)
        }
        if self.dispatcher.log_error:
            error_object["error"] = traceback.format_exc()
        return (error_object, status_code)



from bclib.db_manager.db_manager import DbManager
from bclib.db_manager.sql_db import SqlDb
from bclib.db_manager.sqlite_db import SQLiteDb
from bclib.db_manager.mongo_db import MongoDb
from bclib.db_manager.rabbit_connection import RabbitConnection
from bclib.db_manager.restful_connection import RESTfulConnection



from typing import Any, TYPE_CHECKING

from bclib.utility import HttpMimeTypes
from bclib.listener.http_listener import HttpBaseDataName, HttpBaseDataType
from ..context.request_context import RequestContext

if TYPE_CHECKING:
    from .. import dispatcher


class WebContext(RequestContext):
    def __init__(self, cms_object: dict,  dispatcher: 'dispatcher.IDispatcher') -> None:
        super().__init__(cms_object, dispatcher)
        self.process_async = True
        self.mime = HttpMimeTypes.HTML

    def generate_response(self, result: Any) -> dict:
        ret_val = super().generate_response(result)
        ret_val[HttpBaseDataType.CMS][HttpBaseDataName.CONTENT] = result
        return ret_val



from bclib.utility import DictEx
from ..cache.signaler import factory
from ..cache.cache_manager import CacheManager


class SignalBaseCacheManager(CacheManager):
    """Signal base cache manager"""

    def __init__(self, options: DictEx) -> None:
        super().__init__()
        self._options = options
        signaler_option = options.signaler if options is not None and "signaler" in options else None
        self._signaler = factory.create_signaler(
            signaler_option, self.reset_cache)



"""base class for signaller"""
from abc import ABC
from typing import Callable


class SignalerBase(ABC):
    """base class for signaller"""

    def __init__(self, options: dict, callback: 'Callable[[list[str]], None]') -> None:
        super().__init__()
        self._options = options
        self._callback = callback



import asyncio
from ..dispatcher.socket_dispatcher import RoutingDispatcher
from bclib.listener import Endpoint,  Message, HttpListener


class DevServerDispatcher(RoutingDispatcher):
    def __init__(self, options: dict):
        super().__init__(options)
        self.__listener = HttpListener(
            Endpoint(self.options.server),
            self._on_message_receive_async)

    async def send_message_async(self, message: Message) -> bool:
        """Send message to endpoint"""

        ret_val = asyncio.Future()
        ret_val.set_result(None)
        return ret_val

    def initialize_task(self):
        super().initialize_task()
        self.__listener.initialize_task(self.event_loop)



from typing import Any, TYPE_CHECKING

from bclib.listener.http_listener import HttpBaseDataName, HttpBaseDataType
from bclib.utility import DictEx, HttpStatusCodes, HttpMimeTypes, ResponseTypes
from ..context.context import Context

if TYPE_CHECKING:
    from .. import dispatcher


class RequestContext(Context):
    """Base class for dispatching http base request context"""

    def __init__(self, cms_object: dict,  dispatcher: 'dispatcher.IDispatcher') -> None:
        super().__init__(dispatcher)
        self.cms = DictEx(cms_object)
        self.url: str = self.cms.request.url
        self.query: DictEx = self.cms.query
        self.form: DictEx = self.cms.form
        self.__headers: dict = None
        self.response_type: ResponseTypes = ResponseTypes.RENDERED
        self.status_code: HttpStatusCodes = HttpStatusCodes.OK
        self.mime = HttpMimeTypes.HTML

    def add_header(self, key: str, value: str) -> None:
        """Adding item to response header"""

        if self.__headers is None:
            self.__headers = dict()
        if key not in self.__headers:
            self.__headers[key] = [value]
        else:
            self.__headers[key].append(value)

    def generate_error_response(self, exception: Exception) -> dict:
        """Generate error response from process result"""
        error_object, self.status_code = self._generate_error_object(exception)
        content = f"{error_object['errorMessage']} (Error Code: {error_object['errorCode']})"
        if 'error' in error_object:
            error = error_object["error"].replace("\n", "</br>")
            content += f"<hr/>{error}"
        return self.generate_response(content)

    def generate_response(self, result: Any) -> dict:
        """Generate response from process result"""

        ret_val = self.cms
        ret_val[HttpBaseDataType.CMS][HttpBaseDataName.WEB_SERVER] = {
            "index": self.response_type.value,
            "headercode": self.status_code.value,
            "mime": self.mime
        }
        if self.__headers is not None:
            RequestContext.__add_user_defined_headers(ret_val, self.__headers)
        return ret_val

    @staticmethod
    def __add_user_defined_headers(response: dict, headers: dict) -> None:
        """Adding user defined header to response"""

        if HttpBaseDataType.CMS not in response:
            response[HttpBaseDataType.CMS] = {}
        if HttpBaseDataName.HTTP not in response[HttpBaseDataType.CMS]:
            response[HttpBaseDataType.CMS][HttpBaseDataName.HTTP] = {}
        http = response[HttpBaseDataType.CMS][HttpBaseDataName.HTTP]
        for key, value in headers.items():
            if key in http:
                current_value = http[key] if isinstance(
                    http[key], list) else [http[key]]
                new_value = current_value + value
            else:
                new_value = value

            http[key] = ",".join(new_value)



from bclib.dispatcher.dispatcher import Dispatcher
from bclib.dispatcher.idispatcher import IDispatcher
from bclib.dispatcher.socket_dispatcher import SocketDispatcher
from bclib.dispatcher.dev_server_dispatcher import DevServerDispatcher
from bclib.dispatcher.routing_dispatcher import RoutingDispatcher



from ..cache.cache_manager import CacheManager


class NoCacheManager(CacheManager):
    """"Implementing non caching. Only palace holder for None setting"""

    def cache_decorator(self, seconds: int = 0, key: str = None):
        def decorator(function):
            return function
        return decorator

    def reset_cache(self, keys: 'list[str]'):
        pass

    def get_cache(self, key: str) -> list:
        pass

    def update_cache(self, key: str, data: any) -> bool:
        pass



import json
from pathlib import Path
import multiprocessing

from bclib.dispatcher import RoutingDispatcher, IDispatcher, SocketDispatcher, DevServerDispatcher
from bclib.context import Context, WebContext, SocketContext, ClientSourceContext, ClientSourceMemberContext, RabbitContext, RESTfulContext, RequestContext, MergeType, ServerSourceContext, ServerSourceMemberContext, SourceContext, SourceMemberContext
from bclib.utility import DictEx, HttpStatusCodes, HttpMimeTypes, ResponseTypes, HttpHeaders
from bclib.listener import Message, MessageType, HttpBaseDataType, HttpBaseDataName
from bclib.predicate import Predicate
from bclib.exception import ShortCircuitErr, UnauthorizedErr, HandlerNotFoundErr, InternalServerErr, NotFoundErr
from bclib import __version__


def from_config(option_file_path: str, file_name: str = "host.json"):
    __print_splash(True)
    with open(Path(option_file_path).with_name(file_name), encoding="utf-8") as options_file:
        host_config = json.load(options_file)
    process_list: 'list[multiprocessing.Process]' = list()
    for options in host_config:
        options["__edge_multi_mode__"] = True
        code_file_path = options["code"]
        process = multiprocessing.Process(target=__run_edge_server, args=(
            code_file_path, Path(code_file_path).name, options))
        process_list.append(process)

    for process in process_list:
        process.start()

    for process in process_list:
        process.join()


def __run_edge_server(file_path: str, file_name: str, options: dict()):
    with open(file_path, encoding="utf-8") as file:
        code_str = file.read()
    code = compile(code_str, file_name, 'exec')
    exec(code, {"options": options})


def from_options(options: dict) -> RoutingDispatcher:
    if "__edge_multi_mode__" not in options:
        __print_splash(False)
    ret_val: RoutingDispatcher = None
    if "server" in options:
        ret_val = DevServerDispatcher(options)
    else:
        ret_val = SocketDispatcher(options)
    return ret_val


def __print_splash(inMultiMode: bool):
    print(f'''
______           _                          _____    _            
| ___ \         (_)                        |  ___|  | |           
| |_/ / __ _ ___ _ ___  ___ ___  _ __ ___  | |__  __| | __ _  ___ 
| ___ \/ _` / __| / __|/ __/ _ \| '__/ _ \ |  __|/ _` |/ _` |/ _ \\
| |_/ / (_| \__ \ \__ \ (_| (_) | | |  __/ | |__| (_| | (_| |  __/
\____/ \__,_|___/_|___/\___\___/|_|  \___| \____/\__,_|\__, |\___|
                                                        __/ |     
                                                       |___/      
***********************************
Basiscore Edge

Welcome To BasisCore Ecosystem
Follow us on https://BasisCore.com/
bclib Version : {__version__}
Run in {'multi' if inMultiMode else 'single'} instance mode!
***********************************
(Press CTRL+C to quit)
''')



class HttpMimeTypes:
    AAC = "audio/aac"
    ABW = "application/x-abiword"
    ARC = "application/x-freearc"
    AVI = "video/x-msvideo"
    AZW = "application/vnd.amazon.ebook"
    BIN = "application/octet-stream"
    BMP = "image/bmp"
    BZ = "application/x-bzip"
    BZ2 = "application/x-bzip2"
    CDA = "application/x-cdf"
    CSH = "application/x-csh"
    CSS = "text/css"
    CSV = "text/csv"
    DOC = "application/msword"
    DOCX = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    EOT = "application/vnd.ms-fontobject"
    EPUB = "application/epub+zip"
    GZ = "application/gzip"
    GIF = "image/gif"
    HTM = "text/html"
    HTML = "text/html"
    ICO = "image/vnd.microsoft.icon"
    ICS = "text/calendar"
    JAR = "application/java-archive"
    JPEG = "image/jpeg"
    JPG = "image/jpeg"
    JS = "text/javascript"
    JSON = "application/json"
    JSONLD = "application/ld+json"
    MID = "audio/midi"
    MIDI = "audio/midi"
    X_MIDI = "audio/x-midi"
    MJS = "text/javascript"
    MP3 = "audio/mpeg"
    MP4 = "video/mp4"
    MPEG = "video/mpeg"
    MPKG = "application/vnd.apple.installer+xml"
    ODP = "application/vnd.oasis.opendocument.presentation"
    ODS = "application/vnd.oasis.opendocument.spreadsheet"
    ODT = "application/vnd.oasis.opendocument.text"
    OGA = "audio/ogg"
    OGV = "video/ogg"
    OGX = "application/ogg"
    OPUS = "audio/opus"
    OTF = "font/otf"
    PNG = "image/png"
    PDF = "application/pdf"
    PHP = "application/x-httpd-php"
    PPT = "application/vnd.ms-powerpoint"
    PPTX = "application/vnd.openxmlformats-officedocument.presentationml.presentation"
    COMPRESS_RAR = "application/vnd.rar"
    RTF = "application/rtf"
    SH = "application/x-sh"
    SVG = "image/svg+xml"
    SWF = "application/x-shockwave-flash"
    TAR = "application/x-tar"
    TIF = "image/tiff"
    TIFF = "image/tiff"
    TS = "video/mp2t"
    TTF = "font/ttf"
    TXT = "text/plain"
    VSD = "application/vnd.visio"
    WAV = "audio/wav"
    WEBA = "audio/webm"
    WEBM = "video/webm"
    WEBP = "image/webp"
    WOFF = "font/woff"
    WOFF2 = "font/woff2"
    XHTML = "application/xhtml+xml"
    XLS = "application/vnd.ms-excel"
    XLSX = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    XML = "application/xml"
    XUL = "application/vnd.mozilla.xul+xml"
    COMPRESS_ZIP = "application/zip"
    VIDEO_3GP = "video/3gpp"
    AUDIO_3GP = "audio/3gpp"
    VIDEO_3G2 = "video/3gpp2"
    AUDIO_3G2 = "audio/3gpp2"
    COMPRESS_7Z = "application/x-7z-compressed"




from typing import Any, Callable, Coroutine

from bclib.predicate import Predicate, InList, Equal, Url, Between, NotEqual, GreaterThan, LessThan, LessThanEqual, GreaterThanEqual, Match, HasValue, Callback, All
from bclib import predicate
from bclib.context import Context


class DispatcherHelper:

    @staticmethod
    def in_list(expression: str, *items) -> Predicate:
        """Create list cheking predicate"""

        return InList(expression,  *items)

    @staticmethod
    def equal(expression: str, value: Any) -> Predicate:
        """Create equality cheking predicate"""

        return Equal(expression, value)

    @staticmethod
    def url(pattern: str) -> Predicate:
        """Create url cheking predicate"""

        return Url(pattern)

    @staticmethod
    def between(expression: str, min_value: int, max_value: int) -> Predicate:
        """Create between cheking predicate"""

        return Between(expression, min_value, max_value)

    @staticmethod
    def not_equal(expression: str, value: Any) -> Predicate:
        """Create not equality cheking predicate"""

        return NotEqual(expression, value)

    @staticmethod
    def greater_than(expression: str, value: int) -> Predicate:
        """Create not greater than cheking predicate"""

        return GreaterThan(expression, value)

    @staticmethod
    def less_than(expression: str, value: int) -> Predicate:
        """Create not less than cheking predicate"""

        return LessThan(expression, value)

    @staticmethod
    def less_than_equal(expression: str, value: int) -> Predicate:
        """Create not less than and equal cheking predicate"""

        return LessThanEqual(expression, value)

    @staticmethod
    def greater_than_equal(expression: str, value: int) -> Predicate:
        """Create not less than and equal cheking predicate"""

        return GreaterThanEqual(expression, value)

    @staticmethod
    def match(expression: str, value: str) -> Predicate:
        """Create regex matching cheking predicate"""

        return Match(expression, value)

    @staticmethod
    def has_value(expression: str) -> Predicate:
        """Create has value cheking predicate"""

        return HasValue(expression)

    @staticmethod
    def callback(callback: 'Callable[[Context],Coroutine[bool]]') -> Predicate:
        """Create Callback cheking predicate"""

        return Callback(callback)

    @staticmethod
    def is_post() -> Predicate:
        """create http post request cheking predicate"""

        return Equal("context.cms.request.methode", "post")

    @staticmethod
    def is_get() -> Predicate:
        """create http get request cheking predicate"""

        return Equal("context.cms.request.methode", "get")

    @staticmethod
    def is_put() -> Predicate:
        """create http put request cheking predicate"""

        return Equal("context.cms.request.methode", "put")

    @staticmethod
    def is_delete() -> Predicate:
        """create http delete request cheking predicate"""

        return Equal("context.cms.request.methode", "delete")

    @staticmethod
    def all(*predicates: 'Predicate') -> Predicate:
        """create all cheking predicate"""

        return All(predicates)

    @staticmethod
    def any(*predicates: 'Predicate') -> Predicate:
        """create any cheking predicate"""

        return predicate.Any(predicates)

    @staticmethod
    def post(url: str) -> Predicate:
        """create http post request cheking predicate"""

        return All(DispatcherHelper.url(url), DispatcherHelper.is_post())

    @staticmethod
    def get(url: str) -> Predicate:
        """create http get request cheking predicate"""

        return All(DispatcherHelper.url(url), DispatcherHelper.is_get())

    @staticmethod
    def put(url: str) -> Predicate:
        """create http put request cheking predicate"""

        return All(DispatcherHelper.url(url), DispatcherHelper.is_put())

    @staticmethod
    def delete(url: str) -> Predicate:
        """create http delete request cheking predicate"""

        return All(DispatcherHelper.url(url), DispatcherHelper.is_delete())



from abc import abstractmethod
from bclib.utility import DictEx

from ..logger.log_schema import LogSchema
from ..logger.ilogger import ILogger


class SchemaBaseLogger(ILogger):

    def __init__(self, options: DictEx) -> None:
        super().__init__()
        self.options = options
        if options.has("url"):
            self.__get_url = options.url
        elif options.has("get_url"):
            self.__get_url = options.get_url
        else:
            raise Exception(
                "url part of schema logger not set. set 'url' or 'get_url'")
        self.__schemas: 'dict[str,dict]' = dict()

    async def __load_schema_async(self, schema_id: int) -> LogSchema:
        import aiohttp
        async with aiohttp.ClientSession() as session:
            async with session.get(self.__get_url, params={"schemaId": schema_id}) as response:
                return LogSchema(await response.json())

    @abstractmethod
    async def _save_schema_async(self, schema: dict):
        """Save scheme async"""

    async def __get_dict_async(self, schema_id: int) -> LogSchema:
        if schema_id not in self.__schemas:
            schema = await self.__load_schema_async(schema_id)
            self.__schemas[schema.schema_id] = schema
        return self.__schemas[schema_id]

    async def log_async(self,  **kwargs):
        """log data async"""
        try:
            if 'schema_id' in kwargs:
                schema_id = kwargs["schema_id"]
                questions = await self.__get_dict_async(schema_id)
                answer = questions.get_answer(kwargs)
                await self._save_schema_async(answer)
            else:
                raise Exception("'schema_id' not set for apply logging!")
        except Exception as ex:
            print(
                f"Error in log with schema logger: {repr(ex)}")



from bclib.utility import DictEx
from ..logger.rabbit_schema_base_logger import RabbitSchemaBaseLogger
from ..logger.restful_schema_base_logger import RESTfulSchemaBaseLogger
from ..logger.no_logger import NoLogger
from ..logger.ilogger import ILogger


class LoggerFactory:

    @staticmethod
    def create(options: DictEx) -> ILogger:
        logger: ILogger = None
        if not options.has("logger"):
            logger = NoLogger()
        else:
            logger_option: DictEx = options.logger
            if not logger_option.has('type'):
                raise Exception("Type property not set for logger!")
            else:
                logger_type = logger_option.type.lower()
                if logger_type == 'schema.restful':
                    logger = RESTfulSchemaBaseLogger(logger_option)
                elif logger_type == "schema.rabbit":
                    logger = RabbitSchemaBaseLogger(logger_option)
                else:
                    raise Exception(
                        f"Type '{logger_type}' not support for logger")
            print(f'{logger.__class__.__name__} start logging')
        return logger



from bclib.utility.http_status_codes import HttpStatusCodes
from ..exception.short_circuit_err import ShortCircuitErr


class InternalServerErr(ShortCircuitErr):
    def __init__(self, message: str = None):
        super().__init__(HttpStatusCodes.INTERNAL_SERVER_ERROR, 'http-500', message)



from bclib.listener.endpoint import Endpoint
from bclib.listener.socket_listener import SocketListener
from bclib.listener.rabbit_bus_listener import RabbitBusListener
from bclib.listener.message import Message
from bclib.listener.message_type import MessageType
from bclib.listener.http_listener.http_listener import HttpListener
from bclib.listener.http_listener.http_base_data_name import HttpBaseDataName
from bclib.listener.http_listener.http_base_data_type import HttpBaseDataType



from struct import error
from bclib.context import RabbitContext
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .. import dispatcher
from ..listener.rabbit_listener import RabbitListener
from bclib.utility import DictEx


class RabbitBusListener(RabbitListener):

    def __init__(self, rabbit_options: DictEx,  dispatcher: 'dispatcher.IDispatcher') -> None:
        super().__init__(rabbit_options)
        self.__dispatcher = dispatcher

    def on_rabbit_message_received(self, body):
        try:
            message = {
                "host": self._host,
                "queue": self._queue_name,
                "message":  body.decode("utf-8")
            }
            new_context = RabbitContext(
                DictEx(message), self.__dispatcher)
            self.__dispatcher.dispatch_async(new_context)
        except error as ex:
            print(
                f"error in dispatcher received message from rabbit in {self._host}:{self._queue_name} ({ex})")



from typing import TYPE_CHECKING

from bclib.parser import HtmlParserEx
from bclib.utility import DictEx
from ..context.json_base_request_context import JsonBaseRequestContext

if TYPE_CHECKING:
    from .. import dispatcher


class ClientSourceContext(JsonBaseRequestContext):
    """Context for client dbSource request"""

    def __init__(self, cms_object: dict, dispatcher: 'dispatcher.IDispatcher') -> None:
        super().__init__(cms_object, dispatcher)
        parser = HtmlParserEx()
        self.raw_command = self.cms.form.command
        self.dmn_id = self.cms.form.dmnid if self.cms.form.has(
            'dmnid') else None
        parser.feed(self.raw_command)
        self.command = parser.get_dict_ex()
        self.params: DictEx = None
        if "params" in self.command:
            if len(self.command.params) > 0 and "add" in self.command.params[0]:
                tmp_dic = dict()
                for item in self.command.params[0].add:
                    tmp_dic[item.name] = item.value
                self.params = DictEx(tmp_dic)
        self.process_async = True



import threading
from ..dispatcher.routing_dispatcher import RoutingDispatcher
from ..listener import Endpoint, Message, SocketListener


class SocketDispatcher(RoutingDispatcher):
    def __init__(self, options: dict):
        super().__init__(options)
        self.__lock = threading.Lock()
        self.__listener = SocketListener(
            Endpoint(self.options.receiver),
            Endpoint(self.options.sender),
            self._on_message_receive_async)

    async def send_message_async(self, message: Message) -> bool:
        """Send message to endpoint"""

        try:
            self.__lock.acquire()
            return await self.__listener.send_message_async(message)
        finally:
            self.__lock.release()

    def initialize_task(self):
        super().initialize_task()
        self.__listener.initialize_task(self.event_loop)



import json
from typing import TYPE_CHECKING

from bclib.utility import HttpMimeTypes
from bclib.listener.http_listener import HttpBaseDataName, HttpBaseDataType
from ..context.request_context import RequestContext

if TYPE_CHECKING:
    from .. import dispatcher


class JsonBaseRequestContext(RequestContext):
    """Base class for dispatching http json base request context"""

    def __init__(self, cms_object: dict,  dispatcher: 'dispatcher.IDispatcher') -> None:
        super().__init__(cms_object, dispatcher)
        self.mime = HttpMimeTypes.JSON

    def generate_error_response(self,  exception: Exception) -> dict:
        """Generate error response from process result"""
        error_object, self.status_code = self._generate_error_object(exception)
        return self.generate_response(error_object)

    def generate_response(self, result: dict) -> dict:
        """Generate response from process result"""

        ret_val = super().generate_response(result)
        ret_val[HttpBaseDataType.CMS][HttpBaseDataName.CONTENT] = json.dumps(
            result)
        return ret_val



from bclib.context import Context
from ..predicate.predicate import Predicate
from typing import Any


class HasValue (Predicate):
    """Create has value cheking predicate"""

    def __init__(self, expression: str) -> None:
        super().__init__(expression)

    async def check_async(self, context: Context) -> bool:
        try:
            value = eval(self.exprossion, {}, {"context": context})
            return False if not value or value.isspace() else True
        except:  # pylint: disable=bare-except
            return False



"""RESTful implementation of Db wrapper"""
from typing import Any
from ..db_manager.db import Db


class RESTfulConnection(Db):
    """RESTful implementation of Db wrapper"""

    def __init__(self, connection_string: str) -> None:
        super().__init__()
        self.__api_url = connection_string

    async def get_async(self, segment: str = None, params: dict = None) -> Any:
        """Send get request to web api"""

        url = self.__api_url if segment is None else '{0}{1}'.format(
            self.__api_url, segment)
        import aiohttp
        async with aiohttp.ClientSession() as session:
            async with session.get(url, params=params) as response:
                return await response.json()

    async def post_async(self, segment: str = None, params: dict = None) -> Any:
        """Send post request to web api"""

        url = self.__api_url if segment is None else '{0}{1}'.format(
            self.__api_url, segment)
        import aiohttp
        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=params) as response:
                return await response.json()



from typing import Any
from bclib.parser.html.html_parser_ex import HtmlParserEx
from bclib.parser.answer import Answer, UserActionTypes


def ParseAnswer(json: 'str|Any') -> Answer:
    return Answer(json)



import asyncio
import json
from ..logger.schema_base_logger import SchemaBaseLogger
from bclib.utility import DictEx


class RabbitSchemaBaseLogger(SchemaBaseLogger):
    def __init__(self, options: DictEx) -> None:
        super().__init__(options)
        if not options.has("connection") or not options.connection.has("url") or not options.connection.has("queue"):
            raise Exception(
                "connection part of schema logger not set.")

    async def _save_schema_async(self, schema: dict):
        def send_to_rabbit(options):
            import pika
            connection = pika.BlockingConnection(
                pika.URLParameters(options.connection.url))
            channel = connection.channel()
            channel.queue_declare(queue=options.connection.queue)
            channel = connection.channel()
            channel.basic_publish(
                exchange='', routing_key=options.connection.queue, body=json.dumps(schema))
        loop = asyncio.get_running_loop()
        future = loop.run_in_executor(None, send_to_rabbit, self.options)
        await future



from struct import error
from typing import Callable
import json
import asyncio
from ..signaler.signaler_base import SignalerBase


class RabbitSignaller(SignalerBase):
    """Implement rabbit-mq signaler"""

    def __init__(self, options: dict, callback: 'Callable[[list], None]') -> None:
        super().__init__(options, callback)
        import pika
        try:
            param = pika.URLParameters(options.url)
            queue_name = options.queue
            connection = pika.BlockingConnection(param)
            channel = connection.channel()
            channel.queue_declare(queue=queue_name)

            def on_rabbit_message_received(channel, method, properties, body):
                print(
                    f"Message received from {param.host}:{queue_name} ({body})")
                try:
                    cmd = json.loads(body)
                    if cmd and "type" in cmd:
                        cmd_type = cmd["type"]
                        if cmd_type == "clear-cache" and "keys" in cmd:
                            keys = cmd["keys"]
                            self._callback(keys)

                except error as ex:
                    print(
                        f"error in process received message from rabbit in {param.host}:{queue_name} ({ex})")

            channel.basic_consume(
                queue=queue_name, on_message_callback=on_rabbit_message_received, auto_ack=True)

            print(f'Waiting for messages from {param.host}:{queue_name}.')
            loop = asyncio.get_running_loop()
            loop.run_in_executor(None, channel.start_consuming)
        except Exception as ex:
            print(f"Error in config rabbit-mq ({ex})")
            raise ex



from bclib.utility import DictEx
from ..cache.cache_manager import CacheManager
from ..cache.in_memory_cache_manager import InMemoryCacheManager
from ..cache.no_cache import NoCacheManager


def create_chaching(options: DictEx) -> CacheManager:
    ret_val: CacheManager = None
    cache_type = options.type if options is not None and "type" in options else "none"
    if cache_type == "memory":
        ret_val = InMemoryCacheManager(options)
    elif cache_type == "none":
        ret_val = NoCacheManager()
    else:
        raise Exception(
            f"unknown type for cache ('${cache_type}')")
    return ret_val



"""Base class for implement data base interface"""
from abc import ABC


class Db(ABC):
    """Base class for implement data base interface"""

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



import asyncio
from abc import ABC, abstractmethod
from bclib.utility import DictEx


class RabbitListener(ABC):

    def __init__(self, connection_options: DictEx) -> None:
        import pika
        try:
            param = pika.URLParameters(connection_options.url)
            self._host: str = param.host
            self._queue_name: str = connection_options.queue
            self.__connection = pika.BlockingConnection(param)
            self.__channel = self.__connection.channel()
            self.__channel.queue_declare(queue=self._queue_name)

            self.__channel.basic_consume(
                queue=self._queue_name, on_message_callback=lambda channel, method, properties, body: self.on_rabbit_message_received(body), auto_ack=True)
        except Exception as ex:
            print(f"Error in config rabbit-mq ({ex})")
            raise ex

    @abstractmethod
    def on_rabbit_message_received(self, body):
        pass

    def initialize_task(self, loop: asyncio.AbstractEventLoop):
        loop.create_task(self.__consuming_task())

    async def __consuming_task(self):
        try:
            print(
                f'Waiting for messages from {self._host}:{self._queue_name}.')
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(None, self.__channel.start_consuming)
        except asyncio.CancelledError:
            self.__channel.stop_consuming()
            self.__channel.cancel()
            try:
                self.__channel.close()
            except:
                pass
            try:
                self.__connection.close()
            except:
                pass



from ..logger.schema_base_logger import SchemaBaseLogger
from bclib.utility import DictEx


class RESTfulSchemaBaseLogger(SchemaBaseLogger):
    def __init__(self, options: DictEx) -> None:
        super().__init__(options)
        if options.has("url"):
            self.__post_url = options.url
        elif options.has("post_url"):
            self.__post_url = options.post_url
        else:
            raise Exception(
                "url part of schema logger not set. set 'url' or 'post_url'")

    async def _save_schema_async(self, schema: dict):
        import aiohttp
        async with aiohttp.ClientSession() as session:
            async with session.post(self.__post_url, json={"schema": schema}) as response:
                if response.status != 200:
                    raise Exception(
                        f"Error in send answer schema to server.(status code ={response.status}) :{await response.text()}")



from typing import Any, Callable, Coroutine
from ..context import Context
from ..predicate import Predicate
from bclib.exception import ShortCircuitErr


class CallbackInfo:
    def __init__(self, predicates: 'list[Predicate]',  async_callback: 'Callable[[Context], Coroutine[Any]]') -> Any:
        self.__async_callback = async_callback
        self.__predicates = predicates

    async def try_execute_async(self, context: Context) -> Any:
        result: Any = None
        for predicate in self.__predicates:
            try:
                if not (await predicate.check_async(context)):
                    break
            except ShortCircuitErr as ex:
                result = context.generate_error_response(ex)
                break
        else:
            result = await self.__async_callback(context)
        return result



from bclib.utility import HttpStatusCodes


class ShortCircuitErr(Exception):
    def __init__(self, status_code: HttpStatusCodes, error_code: str = -1, message: str = None):
        super().__init__(message if message else '')
        self.status_code = status_code
        self.error_code = error_code




"""Represent no chanching"""
from abc import ABC, abstractmethod


class CacheManager(ABC):
    """Represent no chanching"""

    @abstractmethod
    def cache_decorator(self, seconds: int = 0, key: str = None):
        """Cache result of function for seconds of time or until signal by key for clear"""

    @abstractmethod
    def reset_cache(self, keys: 'list[str]'):
        """Remove key related cache"""

    @abstractmethod
    def get_cache(self, key: str) -> list:
        """Get key related cached data"""

    @abstractmethod
    def update_cache(self, key: str, data: any) -> bool:
        """Update key related cached data"""



from bclib.utility import DictEx
from ..db_manager.rabbit_connection import RabbitConnection
from ..db_manager.db import Db
from ..db_manager.mongo_db import MongoDb
from ..db_manager.sql_db import SqlDb
from ..db_manager.sqlite_db import SQLiteDb
from ..db_manager.restful_connection import RESTfulConnection


class DbManager:

    def __init__(self, options: DictEx) -> None:
        self._options = options
        self._connections: dict(str, list) = dict()
        settings = options.settings if "settings" in options else None
        if settings:
            for k, setting in [(k.split(".", 2)[1:], v) for k, v in settings.items() if k.find("connections.") == 0]:
                db_type = k[0].lower()
                name = k[1].lower()
                self._connections[name] = [db_type, setting]

    def open_connection(self, key: str) -> Db:
        ret_val: Db = None
        try:
            data = self._connections[key]
        except KeyError as ex:
            raise Exception(
                f"Connection setting with name '{key}' not found!") from ex
        db_type = data[0]
        setting = data[1]
        if db_type == "sql":
            ret_val = SqlDb(data[1])
        elif db_type == "sqlite":
            ret_val = SQLiteDb(setting)
        elif db_type == "mongo":
            ret_val = MongoDb(setting)
        elif db_type == "rest":
            ret_val = RESTfulConnection(setting)
        elif db_type == "rabbit":
            ret_val = RabbitConnection(setting)
        else:
            print(
                f"Data base of type '{db_type}' not supported in this vestion")
        return ret_val

    def open_sql_connection(self, key: str) -> SqlDb:
        return self.open_connection(key)

    def open_sqllite_connection(self, key: str) -> SQLiteDb:
        return self.open_connection(key)

    def open_mongo_connection(self, key: str) -> MongoDb:
        return self.open_connection(key)

    def open_restful_connection(self, key: str) -> RESTfulConnection:
        return self.open_connection(key)

    def open_rabbit_connection(self, key: str) -> RabbitConnection:
        return self.open_connection(key)



import json
from typing import Any, Callable
from bclib import edge, db_manager
from .user_action_types import UserActionTypes
from .user_action import UserAction


class Answer:
    """BasisJsonParser is a tool to parse basis_core components json objects. This tool is developed based on
             basis_core key and values."""

    def __init__(self, data: 'str|Any', api_url: 'str' = None):
        self.json = json.loads(data) if isinstance(data, str) else data
        self.__answer_list: 'list[UserAction]' = None
        self.__api_connection = db_manager.RESTfulConnection(
            api_url) if api_url else None

    async def __fill_answer_list_async(self):
        self.__answer_list = list()
        for data in self.json['properties']:
            for action_type in UserActionTypes:
                if action_type.value in list(data.keys()):
                    for actions in data[action_type.value]:
                        if 'parts' in actions.keys():
                            for parts in actions['parts']:
                                for values in parts['values']:
                                    prp_id = data['propId']
                                    prp_value_id = actions['id'] if 'id' in actions.keys(
                                    ) else None
                                    part_number = parts['part'] if "part" in parts.keys(
                                    ) else None
                                    value_id = values['id'] if "id" in values.keys(
                                    ) else None
                                    value = values['value']
                                    self.__answer_list.append(UserAction(
                                        prp_id, action_type, prp_value_id, value_id, value, None, None, part_number))
                        else:
                            prp_id = data['propId']
                            prp_value_id = actions['id'] if 'id' in actions.keys(
                            ) else None
                            self.__answer_list.append(UserAction(
                                prp_id, action_type,  prp_value_id, None, None, None, None, None))

        await self.__try_set_data_type_async()

    async def __get_action_async(self, prp_id_list: 'list[int]', action_list: 'list[UserActionTypes]', part_list: 'list[int]', predicate: 'Callable[[UserAction],bool]' = None) -> 'list[UserAction]':
        ret_val: 'list[UserAction]' = None
        if self.__answer_list is None:
            await self.__fill_answer_list_async()

        if predicate:
            ret_val = [x for x in self.__answer_list if predicate(x)]
        else:
            ret_val = [x for x in self.__answer_list if
                       (prp_id_list is None or x.prp_id in prp_id_list) and
                       (action_list is None or x.action in action_list) and
                       (part_list is None or x.part in part_list)
                       ]
        return ret_val if ret_val else list()

    async def get_actions_async(self, prp_id: 'int|list[int]' = None, action: 'UserActionTypes|list[UserActionTypes]' = None,
                                part: int = None, predicate: 'Callable[[UserAction],bool]' = None) -> 'list[UserAction]':
        """
        inputs:
        prpid: None, int or list
        action: None, int or list
        part: None or in
        Samples:
        (prpid=None, action='edited')
        (prpid=[12345, 1000], action=None)
        """

        action_list = [action] if isinstance(
            action, UserActionTypes) else action
        prp_id_list = [prp_id] if isinstance(prp_id, int) else prp_id
        part_list = [part] if isinstance(part, int) else part
        return await self.__get_action_async(prp_id_list, action_list, part_list, predicate)

    def __data_type_checker(self, view_type: str, datatype: str = None):

        if view_type in ["select", "checkList"]:
            result = "fixvalue"
        elif view_type == "textarea":
            result = "ntextvalue"
        elif view_type == "text" and datatype in ["text", "None", None]:
            result = "textvalue"
        elif view_type == "text" and datatype == "int":
            result = "numvalue"
        elif view_type == "text" and datatype == "float":
            result = "floatvalue"
        elif view_type == "autocomplete":
            result = "autocomplete"
        else:
            result = "None"
        return result

    async def __try_set_data_type_async(self):
        if self.__api_connection:
            type_list = list()
            questions_data = edge.DictEx(await self.__api_connection.post_async())
            for data in questions_data.sources:
                for question in data.data:
                    for parts in question.questions:
                        for validations in parts.parts:
                            data_type = validations.validations["datatype"] if "datatype" in validations.validations.keys(
                            ) else "None"
                            type_list.append({
                                "prpId": parts.prpId, "part": validations.part, "viewtype": validations.viewType,
                                "datatype": data_type, "table": self.__data_type_checker(validations.viewType, data_type)
                            })
            for type in type_list:
                for values in self.__answer_list:
                    if int(type['prpId']) == int(values.prp_id) and type["part"] == values.part or values.part is None:
                        values.datatype = type['table']



from bclib.listener.http_listener.http_listener import HttpListener
from bclib.listener.http_listener.http_base_data_name import HttpBaseDataName
from bclib.listener.http_listener.http_base_data_type import HttpBaseDataType



from typing import TYPE_CHECKING
from bclib.utility import DictEx
from .context import Context

if TYPE_CHECKING:
    from .. import dispatcher
    from .. import listener


class SocketContext(Context):
    """Base class for dispatching web socket base request context"""

    def __init__(self, cms_object: dict,  dispatcher: 'dispatcher.IDispatcher', message_object: 'listener.Message', body: dict) -> None:
        super().__init__(dispatcher)
        self.cms = DictEx(cms_object) if cms_object else None
        self.url: str = self.cms.request.url if cms_object else None
        self.body = DictEx(body) if body else None
        self.message_object = message_object
        self.session_id = self.message_object.session_id
        self.message_type = self.message_object.type
        self.is_adhoc = False



from bclib.exception import ShortCircuitErr
from ..predicate.predicate import Predicate
from typing import Callable, Coroutine
from bclib.context import Context


class Callback (Predicate):
    """Create callback base cheking predicate"""

    def __init__(self, callback: 'Callable[[Context],Coroutine[bool]]') -> None:
        super().__init__(None)
        self.__callback = callback

    async def check_async(self, context: Context) -> bool:
        try:
            return await self.__callback(context)
        except ShortCircuitErr:
            raise
        except:  # pylint: disable=bare-except
            return False




from datetime import timedelta
from datetime import datetime
from functools import wraps
from typing import Callable
from bclib.utility import DictEx
from ..cache.signal_base_cache_manager import SignalBaseCacheManager


class InMemoryCacheManager(SignalBaseCacheManager):
    """Implement in memory cache manager"""

    def __init__(self, options: DictEx) -> None:
        super().__init__(options)
        self.__cache_list: 'dict[str, list[Callable]]' = dict()

    def cache_decorator(self, seconds: int = 0, key: str = None):
        """Cache result of function for seconds of time or until signal by key for clear"""

        def decorator(function):
            function.cache = None
            if seconds > 0:
                function.lifetime = timedelta(seconds=seconds)
                function.expiration = datetime.utcnow() + function.lifetime
            if key is not None:
                if key not in self.__cache_list:
                    self.__cache_list[key] = list()
                self.__cache_list[key].append(function)

            @wraps(function)
            def wrapper_with_time(*args, **kwargs):
                if function.cache is not None and datetime.utcnow() >= function.expiration:
                    function.cache = None
                    function.expiration = datetime.utcnow() + function.lifetime
                if function.cache is None:
                    function.cache = function(*args, **kwargs)
                return function.cache

            @wraps(function)
            def wrapper_without_time(*args, **kwargs):
                if function.cache is None:
                    function.cache = function(*args, **kwargs)
                return function.cache
            return wrapper_with_time if seconds > 0 else wrapper_without_time
        return decorator

    def reset_cache(self, keys: 'list[str]') -> None:
        """Remove key related cache"""

        print(f"reset cache for {keys}")
        for key in keys:
            if key in self.__cache_list:
                for function in self.__cache_list[key]:
                    function.cache = None

    def get_cache(self, key: str) -> list:
        """Get key related cached data"""

        return [function.cache for function in self.__cache_list[key]]

    def update_cache(self, key: str, data: any) -> bool:
        """Update key related cached data"""

        is_siccessfull = False
        if key in self.__cache_list:
            for function in self.__cache_list[key]:
                function.cache = data
                is_siccessfull = True
        return is_siccessfull



from typing import Callable
from ..signaler.no_signaler import NoSignaler
from ..signaler.signaler_base import SignalerBase
from ..signaler.rabbit_signaler import RabbitSignaller


def create_signaler(options: dict, callback: 'Callable[[list[str]], None]') -> SignalerBase:
    ret_val: None
    signaler_type = options["type"] if options and "type" in options else "none"
    if signaler_type == "rabbit":
        ret_val = RabbitSignaller(options, callback)
    elif signaler_type == "none":
        ret_val = NoSignaler()
    else:
        raise Exception(
            f"unknown type for cache signaller('${signaler_type}')")

    return ret_val



"""Restful implementation of Db wrapper"""
import json
from typing import Any
from bclib.utility import DictEx
from ..db_manager.db import Db


class RabbitConnection(Db):
    """Restful implementation of Db wrapper"""

    def __init__(self, connection_setting: DictEx) -> None:
        super().__init__()
        self.__connection_setting = connection_setting
        self.__connection = None
        self.__channel = None

    def __enter__(self):
        import pika
        connection = pika.BlockingConnection(
            pika.URLParameters(self.__connection_setting.host))
        self.__channel = connection.channel()
        self.__channel.queue_declare(queue=self.__connection_setting.queue)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.__connection is not None:
            self.__connection.close()
        return super().__exit__(exc_type, exc_val, exc_tb)

    def publish(self, message: Any):
        """Send message to rabbit-mq"""
        if self.__channel is not None:
            msg_json = json.dumps(message)
            self.__channel.basic_publish(
                exchange='', routing_key=self.__connection_setting.queue, body=msg_json)
        else:
            raise Exception(f'for host {0}:{1}, channel not created!' %
                            self.__connection_setting.host, self.__connection_setting.queue)



"""import html parser"""
from html.parser import HTMLParser
from typing import Any
from bclib.utility import DictEx
from ..html.html_tag import HtmlTag


class HtmlParserEx(HTMLParser):
    """custom parser for row html"""

    def __init__(self,) -> None:
        super().__init__()
        self.element = None
        self.stack = []

    def handle_starttag(self, tag, attrs):
        if self.element is not None:
            self.stack.append(self.element)

        self.element = HtmlTag(tag)
        for name, value in attrs:
            self.element.attributes[name] = value

    def handle_endtag(self, tag):
        if len(self.stack) != 0:
            tmp = self.element
            self.element = self.stack.pop()
            self.element.child_list.append(tmp)

    def handle_data(self, data):
        pass

    def error(self, message: str) -> Any:
        print("Error in parse html", message)
        return super().error(message)

    def get_dict_ex(self) -> DictEx:
        def _convert_tag_to_dict(tag: HtmlTag) -> dict:
            dic = dict(tag.attributes)
            for child in tag.child_list:
                if child.name in dic:
                    dic_list = dic.get(child.name)
                else:
                    dic_list = dic[child.name] = list()
                child_dict = _convert_tag_to_dict(child)
                dic_list.append(child_dict)
            return dic
        new_dic = _convert_tag_to_dict(self.element)
        return DictEx(new_dic)
